from __future__ import annotations

import dataclasses
from typing import TYPE_CHECKING, Any, List, Sequence, Set, Type, Union

import pyarrow
from pydantic import BaseModel
from typing_extensions import TypeAlias

from chalk.features import DataFrame
from chalk.utils import AnyDataclass

if TYPE_CHECKING:
    import polars as pl


@dataclasses.dataclass
class StreamResolverParam:
    name: str


@dataclasses.dataclass
class StreamResolverParamMessage(StreamResolverParam):
    typ: Union[Type[str], Type[bytes], Type[BaseModel], AnyDataclass]


StreamResolverWindowType: TypeAlias = Union[
    Type[List[str]],
    Type[List[bytes]],
    Type[List[BaseModel]],
    Type[AnyDataclass],
    Type[pyarrow.Table],
    Type["pl.DataFrame"],
    Type[DataFrame],
    Any,  # The annotation value is likely going to be a GenericAlias, which messes with pydantic's validation
]


@dataclasses.dataclass
class StreamResolverParamMessageWindow(StreamResolverParam):
    typ: StreamResolverWindowType


@dataclasses.dataclass
class StreamResolverSignature:
    params: Sequence[StreamResolverParam]
    output_feature_fqns: Set[str]


@dataclasses.dataclass
class StreamResolverParamKeyedState(StreamResolverParam):
    typ: Union[Type[BaseModel], Type[AnyDataclass]]
    default_value: Any



from __future__ import annotations

from typing import TYPE_CHECKING, Callable, List

if TYPE_CHECKING:
    import polars as pl

    from chalk.features import Filter


class Aggregation:
    def __init__(self, col: pl.Expr, fn: Callable[[pl.Expr], pl.Expr]):
        self.col = col
        self.fn = fn
        self.filters: List[Filter] = []

    def where(self, *f: Filter):
        # TODO: Make this support using `and` and `or` between filters.
        #       Right now, the iterable of filters is taken to mean `and`.
        self.filters.extend(f)
        return self

    def __add__(self, other):
        return self

    def __radd__(self, other):
        return self


class op:
    """Operations for aggregations in `DataFrame`.

    The class methods on this class are used to create aggregations
    for use in `DataFrame.group_by`.
    """

    @classmethod
    def sum(cls, col, *cols) -> Aggregation:
        """Add together the values of `col` and `*cols` in a `DataFrame`.

        Parameters
        ----------
        col
            There must be at least one column to aggregate.
        cols
            Subsequent columns to aggregate.

        Examples
        --------
        >>> from chalk.features import DataFrame
        >>> df = DataFrame(
        ...     {
        ...         User.id: [1, 1, 3],
        ...         User.val: [0.5, 4, 10],
        ...     }
        ... ).group_by(
        ...      group={User.id: User.id}
        ...      agg={User.val: op.sum(User.val)}
        ... )
        ╭─────────┬──────────╮
        │ User.id │ User.val │
        ├─────────┼──────────┤
        │ i64     │ f64      │
        ╞═════════╪══════════╡
        │  1      │ 4.5      │
        ├─────────┼──────────┤
        │  3      │ 10       │
        ╰─────────┴──────────╯
        """
        import polars as pl

        cols = [col, *[t for t in cols]]
        if len(cols) == 1:
            return Aggregation(pl.col(str(cols[0])), lambda x: x.sum())
        c = pl.col([str(c) for c in cols])
        return Aggregation(c, lambda x: pl.sum(c))

    @classmethod
    def product(cls, col) -> Aggregation:
        """Multiply together the values of `col` in a `DataFrame`.

        Parameters
        ----------
        col
            The column to aggregate. Used in `DataFrame.group_by`.

        Examples
        --------
        >>> from chalk.features import DataFrame
        >>> df = DataFrame(
        ...     {
        ...         User.id: [1, 1, 3],
        ...         User.val: [0.5, 4, 10],
        ...         User.active: [True, True, False],
        ...     }
        ... ).group_by(
        ...      group={User.id: User.id}
        ...      agg={
        ...         User.val: op.product(User.val),
        ...         User.active: op.product(User.active),
        ...      }
        ... )
        ╭─────────┬──────────┬─────────────╮
        │ User.id │ User.val │ User.active │
        ├─────────┼──────────┼─────────────┤
        │ i64     │ f64      │ i64         │
        ╞═════════╪══════════╪═════════════╡
        │  1      │ 2        │ 1           │
        ├─────────┼──────────┼─────────────┤
        │  3      │ 10       │ 0           │
        ╰─────────┴──────────┴─────────────╯
        """
        import polars as pl

        return Aggregation(pl.col(str(col)), lambda x: x.product())

    @classmethod
    def max(cls, col) -> Aggregation:
        """Find the maximum of the values of `col` in a `DataFrame`.

        Parameters
        ----------
        col
            The column along which to find the maximum value.

        Examples
        --------
        >>> from chalk.features import DataFrame
        >>> df = DataFrame(
        ...     {
        ...         User.id: [1, 1, 3],
        ...         User.val: [0.5, 4, 10],
        ...     }
        ... ).group_by(
        ...      group={User.id: User.id}
        ...      agg={User.val: op.max(User.val)}
        ... )
        ╭─────────┬──────────╮
        │ User.id │ User.val │
        ├─────────┼──────────┤
        │ i64     │ f64      │
        ╞═════════╪══════════╡
        │  1      │ 4        │
        ├─────────┼──────────┤
        │  3      │ 10       │
        ╰─────────┴──────────╯
        """
        import polars as pl

        return Aggregation(pl.col(str(col)), lambda x: x.max())

    @classmethod
    def min(cls, col) -> Aggregation:
        """Find the minimum of the values of `col` in a `DataFrame`.

        Parameters
        ----------
        col
            The column along which to find the minimum value.

        Examples
        --------
        >>> from chalk.features import DataFrame
        >>> df = DataFrame(
        ...     {
        ...         User.id: [1, 1, 3],
        ...         User.val: [0.5, 4, 10],
        ...     }
        ... ).group_by(
        ...      group={User.id: User.id}
        ...      agg={User.val: op.min(User.val)}
        ... )
        ╭─────────┬──────────╮
        │ User.id │ User.val │
        ├─────────┼──────────┤
        │ i64     │ f64      │
        ╞═════════╪══════════╡
        │  1      │ 0.5      │
        ├─────────┼──────────┤
        │  3      │ 10       │
        ╰─────────┴──────────╯
        """
        import polars as pl

        return Aggregation(pl.col(str(col)), lambda x: x.min())

    @classmethod
    def mode(cls, col) -> Aggregation:
        import polars as pl

        return Aggregation(pl.col(str(col)), lambda x: x.mode())

    @classmethod
    def quantile(cls, col, q: float) -> Aggregation:
        import polars as pl

        assert q >= 0 <= 1, f"Quantile must be between 0 and 1, but given {q}"
        return Aggregation(pl.col(str(col)), lambda x: x.quantile(q))

    @classmethod
    def median(cls, col) -> Aggregation:
        import polars as pl

        return Aggregation(pl.col(str(col)), lambda x: x.median())

    @classmethod
    def mean(cls, col) -> Aggregation:
        import polars as pl

        return Aggregation(pl.col(str(col)), lambda x: x.mean())

    @classmethod
    def std(cls, col) -> Aggregation:
        import polars as pl

        return Aggregation(pl.col(str(col)), lambda x: x.std())

    @classmethod
    def variance(cls, col) -> Aggregation:
        import polars as pl

        return Aggregation(pl.col(str(col)), lambda x: x.var())

    @classmethod
    def count(cls, col) -> Aggregation:
        import polars as pl

        return Aggregation(pl.col(str(col)), lambda x: x.count())

    @classmethod
    def concat(cls, col, col2, sep: str = "") -> Aggregation:
        import polars as pl

        c = pl.col([str(col), str(col2)])
        return Aggregation(c, lambda x: pl.concat_str(c, sep))

    @classmethod
    def concat_str(cls, col, col2, sep: str = "") -> Aggregation:
        import polars as pl

        c = pl.col([str(col), str(col2)])
        return Aggregation(c, lambda x: pl.concat_str(c, sep))

    # TODO: Add these back with `.sort(...)` on `DataFrame`.
    # @classmethod
    # def last(cls, col) -> Aggregation:
    #     return Aggregation(col, lambda x: x.last())
    #
    # @classmethod
    # def first(cls, col) -> Aggregation:
    #     return Aggregation(col, lambda x: x.first())



import ast
import inspect
from types import FrameType
from typing import Any, Callable, Dict, Optional, Union

from executing import Source

from chalk.features import Filter
from chalk.features._chalkop import op


class ExecutingException(Exception):
    pass


class RetrievingError(ExecutingException):
    pass


def _get_node_by_frame(frame: FrameType, raise_exc: bool = True) -> Optional[ast.AST]:
    """Get the node by frame, raise errors if possible"""
    exect = Source.executing(frame)

    if exect.node:
        # attach the frame for better exception message
        # (i.e. where ImproperUseError happens)
        exect.node.__frame__ = frame  # type: ignore
        return exect.node

    assert isinstance(exect.source, Source)
    if exect.source.text and exect.source.tree and raise_exc:
        raise RetrievingError(
            (
                "Couldn't retrieve the call node. "
                "This may happen if you're using some other AST magic at the "
                "same time, such as pytest, ipython, macropy, or birdseye."
            )
        )

    return None


def _get_func_frame_and_nodes(condition: Callable[[Optional[ast.AST]], bool]):
    frames = inspect.stack()
    for frame_info in frames[3:]:
        frame = frame_info.frame
        func_node = _get_node_by_frame(frame)
        if condition(func_node):
            # This is the correct getitem frame.
            # It is important that the "slice" isn't an ast.Name,
            # as otherwise it would be impossible to parse the expression
            # ast.Name would imply something like this:
            # def __getitem__(self, item):
            #    return self.df[item]  # <--- item is of type ast.name. We need to go one frame higher!
            return frame, func_node
    raise RuntimeError("Condition not found in stack")


def parse_feature_iter(f):
    func_frame, func_node = _get_func_frame_and_nodes(lambda node: isinstance(node, ast.Call))
    if not isinstance(func_node, ast.Call):
        raise RuntimeError("Could not evaluate function")
    if not isinstance(func_node.func, ast.Name):
        raise RuntimeError("Could not evaluate function")
    func_name = func_node.func.id
    if func_name == "sum":
        return op.sum(f)
    elif func_name == "max":
        return op.max(f)
    elif func_name == "min":
        return op.min(f)

    raise RuntimeError(f"Could not evaluate function {func_name}")


def parse_dataframe_getitem():
    func_frame, func_node = _get_func_frame_and_nodes(
        lambda node: isinstance(node, ast.Subscript) and not isinstance(node.slice, ast.Name)
    )
    assert isinstance(func_node, ast.Subscript)
    slc = func_node.slice
    if isinstance(slc, ast.Index):
        slc = slc.value  # type: ignore
        assert isinstance(slc, ast.expr)
    converted_slice = convert_slice(slc)
    return eval_converted_expr(converted_slice, glbs=func_frame.f_globals, lcls=func_frame.f_locals)


def parse_when() -> Optional[Filter]:
    func_frame, func_node = _get_func_frame_and_nodes(lambda node: isinstance(node, ast.Call))
    assert isinstance(func_node, ast.Call)
    when = next((k for k in func_node.keywords if k.arg == "when"), None)
    when_filter = convert_slice(when.value) if when else None
    assert isinstance(when_filter, ast.expr)
    return (
        eval_converted_expr(when_filter, glbs=func_frame.f_globals, lcls=func_frame.f_locals) if when_filter else None
    )


def _convert_maybe_tuple(slc: ast.expr):
    if isinstance(slc, ast.Tuple):
        return ast.Tuple(
            elts=[_convert_ops(x) for x in slc.elts],
            ctx=slc.ctx,
        )
    else:
        assert isinstance(slc, ast.expr)
        return _convert_ops(slc)


def convert_slice(slc: Union[ast.expr, ast.Index]):
    if isinstance(slc, ast.Index):
        # Index is deprecated in Python 3.9+
        slc = slc.value  # type: ignore
        assert isinstance(slc, ast.expr)
        slc = _convert_maybe_tuple(slc)
        return ast.Index(value=slc)
    return _convert_maybe_tuple(slc)


def eval_converted_expr(expr: ast.AST, glbs: Optional[Dict[str, Any]] = None, lcls: Optional[Dict[str, Any]] = None):
    expr.lineno = 1
    expr.col_offset = 0
    expr.end_lineno = 1
    expr.end_col_offset = 0
    expression = ast.Expression(body=expr)
    ast.fix_missing_locations(expression)
    glbs = dict(glbs or {})  # shallow copy
    # Inject the __CHALK_FILTER__ so the converted "in" and "not in" expressions can be parsed
    glbs["__CHALK_FILTER__"] = Filter
    return eval(compile(expression, filename="<string>", mode="eval"), glbs, lcls)  # nosemgrep: eval-detected


def _convert_ops(stmt: ast.expr):
    """Recursively convert operations so that they can be parsed by the filters"""
    if isinstance(stmt, ast.BoolOp):
        assert len(stmt.values) >= 2, "bool ops need at least two values"
        op: ast.operator
        if isinstance(stmt.op, ast.And):
            op = ast.BitAnd()
        elif isinstance(stmt.op, ast.Or):
            op = ast.BitOr()
        else:
            raise ValueError(f"Invalid op: {stmt.op}")
        values = list(stmt.values)
        ans = _convert_ops(values.pop())
        while len(values) > 0:
            left = values.pop()
            ans = ast.BinOp(
                left=_convert_ops(left),
                op=op,
                right=ans,
            )
        return ans
    if isinstance(stmt, ast.UnaryOp):
        if isinstance(stmt.op, ast.Not):
            return ast.UnaryOp(
                op=ast.Invert(),
                operand=_convert_ops(stmt.operand),
            )
        return stmt
    if isinstance(stmt, ast.Compare):
        if len(stmt.ops) == 1:
            lhs = stmt.left
            assert len(stmt.comparators) == 1
            rhs = stmt.comparators[0]
            compare_op: ast.cmpop = stmt.ops[0]
            # Replace is with == and isnot with !=
            # It doesn't make sense to have identity checks in a dataframe filter

            if isinstance(compare_op, ast.Is):
                return ast.Compare(left=lhs, ops=[ast.Eq()], comparators=[rhs])
            if isinstance(compare_op, ast.IsNot):
                return ast.Compare(left=lhs, ops=[ast.NotEq()], comparators=[rhs])

            if isinstance(compare_op, (ast.In, ast.NotIn)):
                filter_op = "in" if isinstance(compare_op, ast.In) else "not in"
                return ast.Call(
                    func=ast.Name(id="__CHALK_FILTER__", ctx=ast.Load()),
                    args=[
                        stmt.left,
                        ast.Constant(value=filter_op),
                        rhs,
                    ],
                    keywords=[],
                )
        return stmt
    return stmt



import json
import os
from typing import Any, Callable, Optional, TypeVar, overload

T = TypeVar("T")


def has_integration(integration_name: str) -> bool:
    encoded = os.getenv("_CHALK_AVAILABLE_INTEGRATIONS")
    if encoded is not None:
        available = set(json.loads(encoded))
        return integration_name in available
    return False


@overload
def load_integration_variable(
    name: str,
    integration_name: Optional[str],
) -> Optional[str]:
    ...


@overload
def load_integration_variable(name: str, integration_name: Optional[str], parser: Callable[[str], T]) -> Optional[T]:
    ...


def load_integration_variable(
    name: str,
    integration_name: Optional[str],
    parser: Callable[[str], Any] = str,
) -> Optional[Any]:
    value = None
    if integration_name is None:
        value = os.getenv(name)
    elif has_integration(integration_name):
        value = os.getenv(f"{integration_name}_{name}")

    return parser(value) if value is not None else None



from typing import Literal

MissingValueStrategy = Literal[
    # Raise a TypeError if missing values are found
    "error",
    # PROBABLY THIS ONE
    # Coerce missing values to the default value for the feature, if one is defined.
    # Otherwise, raise a TypeError
    "default_or_error",
    # Coerce missing values to the default value for the feature, if one is defined.
    # Otherwise, treat missing values as valid.
    "default_or_allow",
    # Treat missing values as valid values
    "allow",
]



# AUTO-GENERATED FILE. Do not edit.
from typing import Any, Dict, List, Literal, Optional, Tuple, TypeVar, Union

from chalk._monitoring.charts_enums_codegen import GroupByKind
from chalk._monitoring.charts_series_base import ResolverType, SeriesBase
from chalk.features.resolver import ResolverProtocol


class Series(SeriesBase):
    """
    Class describing a series of data in two dimensions, as in a line chart.
    Series should be instantiated with one of the classmethods that specifies
    the metric to be tracked.
    """

    def __new__(cls, *args, **kwargs):
        raise ValueError("Please construct a Series with a metric classmethod")

    @classmethod
    def feature_request_count_metric(cls, name: str) -> "FeatureRequestCountSeries":
        """Creates a Series of metric kind FeatureRequestCount.

        Parameters
        ----------
        name
            A name for your new `feature_request_count` Series.

        Returns
        -------
        FeatureRequestCountSeries
            A new `FeatureRequestCountSeries` instance that inherits from the `Series` class.
        """
        return FeatureRequestCountSeries(name=name, metric="FEATURE_REQUEST_COUNT", window_function="COUNT")

    @classmethod
    def feature_staleness_metric(
        cls,
        name: str,
        window_function: Literal["count", "mean", "sum", "max", "99%", "95%", "75%", "50%", "25%", "5%", "min", "all"],
    ) -> "FeatureStalenessSeries":
        """Creates a Series of metric kind `FeatureStaleness`.

        Parameters
        ----------
        name
            A name for your new `feature_staleness` Series.
        window_function
            The time window to calculate the metric over.

        Returns
        -------
        FeatureStalenessSeries
            A new `FeatureStalenessSeries` instance that inherits from the `Series` class.
        """
        if window_function not in {
            "count",
            "mean",
            "sum",
            "max",
            "99%",
            "95%",
            "75%",
            "50%",
            "25%",
            "5%",
            "min",
            "all",
        }:
            raise ValueError(f"window_function value '{window_function}' is not valid")
        return FeatureStalenessSeries(name=name, metric="FEATURE_STALENESS", window_function=window_function)

    @classmethod
    def feature_value_metric(
        cls,
        name: str,
        window_function: Literal["mean", "sum", "max", "99%", "95%", "75%", "50%", "25%", "5%", "min", "all"],
    ) -> "FeatureValueSeries":
        """Creates a Series of metric kind `FeatureValue`.

        Parameters
        ----------
        name
            A name for your new `feature_value` Series.
        window_function
            The time window to calculate the metric over.

        Returns
        -------
        FeatureValueSeries
            A new `FeatureValueSeries` instance that inherits from the `Series` class.
        """
        if window_function not in {"mean", "sum", "max", "99%", "95%", "75%", "50%", "25%", "5%", "min", "all"}:
            raise ValueError(f"window_function value '{window_function}' is not valid")
        return FeatureValueSeries(name=name, metric="FEATURE_VALUE", window_function=window_function)

    @classmethod
    def feature_null_ratio_metric(cls, name: str) -> "FeatureNullRatioSeries":
        """Creates a Series of metric kind `FeatureNullRatio`.

        Parameters
        ----------
        name
            A name for your new `feature_null_ratio` Series.

        Returns
        -------
        FeatureNullRatioSeries
            A new `FeatureNullRatioSeries` instance that inherits from the `Series` class.
        """
        return FeatureNullRatioSeries(
            name=name,
            metric="FEATURE_NULL_RATIO",
        )

    @classmethod
    def resolver_request_count_metric(cls, name: str) -> "ResolverRequestCountSeries":
        """Creates a Series of metric kind ResolverRequestCount.

        Parameters
        ----------
        name
            A name for your new `resolver_request_count` Series.

        Returns
        -------
        ResolverRequestCountSeries
            A new `ResolverRequestCountSeries` instance that inherits from the `Series` class.
        """
        return ResolverRequestCountSeries(name=name, metric="RESOLVER_REQUEST_COUNT", window_function="COUNT")

    @classmethod
    def resolver_latency_metric(
        cls,
        name: str,
        window_function: Literal["count", "mean", "sum", "max", "99%", "95%", "75%", "50%", "25%", "5%", "min", "all"],
    ) -> "ResolverLatencySeries":
        """Creates a Series of metric kind `ResolverLatency`.

        Parameters
        ----------
        name
            A name for your new `resolver_latency` Series.
        window_function
            The time window to calculate the metric over.

        Returns
        -------
        ResolverLatencySeries
            A new `ResolverLatencySeries` instance that inherits from the `Series` class.
        """
        if window_function not in {
            "count",
            "mean",
            "sum",
            "max",
            "99%",
            "95%",
            "75%",
            "50%",
            "25%",
            "5%",
            "min",
            "all",
        }:
            raise ValueError(f"window_function value '{window_function}' is not valid")
        return ResolverLatencySeries(name=name, metric="RESOLVER_LATENCY", window_function=window_function)

    @classmethod
    def resolver_success_ratio_metric(cls, name: str) -> "ResolverSuccessRatioSeries":
        """Creates a Series of metric kind `ResolverSuccessRatio`.

        Parameters
        ----------
        name
            A name for your new `resolver_success_ratio` Series.

        Returns
        -------
        ResolverSuccessRatioSeries
            A new `ResolverSuccessRatioSeries` instance that inherits from the `Series` class.
        """
        return ResolverSuccessRatioSeries(
            name=name,
            metric="RESOLVER_SUCCESS_RATIO",
        )

    @classmethod
    def query_count_metric(cls, name: str) -> "QueryCountSeries":
        """Creates a Series of metric kind QueryCount.

        Parameters
        ----------
        name
            A name for your new `query_count` Series.

        Returns
        -------
        QueryCountSeries
            A new `QueryCountSeries` instance that inherits from the `Series` class.
        """
        return QueryCountSeries(name=name, metric="QUERY_COUNT", window_function="COUNT")

    @classmethod
    def query_latency_metric(
        cls,
        name: str,
        window_function: Literal["count", "mean", "sum", "max", "99%", "95%", "75%", "50%", "25%", "5%", "min", "all"],
    ) -> "QueryLatencySeries":
        """Creates a Series of metric kind `QueryLatency`.

        Parameters
        ----------
        name
            A name for your new `query_latency` Series.
        window_function
            The time window to calculate the metric over.

        Returns
        -------
        QueryLatencySeries
            A new `QueryLatencySeries` instance that inherits from the `Series` class.
        """
        if window_function not in {
            "count",
            "mean",
            "sum",
            "max",
            "99%",
            "95%",
            "75%",
            "50%",
            "25%",
            "5%",
            "min",
            "all",
        }:
            raise ValueError(f"window_function value '{window_function}' is not valid")
        return QueryLatencySeries(name=name, metric="QUERY_LATENCY", window_function=window_function)

    @classmethod
    def query_success_ratio_metric(cls, name: str) -> "QuerySuccessRatioSeries":
        """Creates a Series of metric kind `QuerySuccessRatio`.

        Parameters
        ----------
        name
            A name for your new `query_success_ratio` Series.

        Returns
        -------
        QuerySuccessRatioSeries
            A new `QuerySuccessRatioSeries` instance that inherits from the `Series` class.
        """
        return QuerySuccessRatioSeries(
            name=name,
            metric="QUERY_SUCCESS_RATIO",
        )

    @classmethod
    def cron_count_metric(cls, name: str) -> "CronCountSeries":
        """Creates a Series of metric kind CronCount.

        Parameters
        ----------
        name
            A name for your new `cron_count` Series.

        Returns
        -------
        CronCountSeries
            A new `CronCountSeries` instance that inherits from the `Series` class.
        """
        return CronCountSeries(name=name, metric="CRON_COUNT", window_function="COUNT")

    @classmethod
    def cron_latency_metric(
        cls,
        name: str,
        window_function: Literal["count", "mean", "sum", "max", "99%", "95%", "75%", "50%", "25%", "5%", "min", "all"],
    ) -> "CronLatencySeries":
        """Creates a Series of metric kind `CronLatency`.

        Parameters
        ----------
        name
            A name for your new `cron_latency` Series.
        window_function
            The time window to calculate the metric over.

        Returns
        -------
        CronLatencySeries
            A new `CronLatencySeries` instance that inherits from the `Series` class.
        """
        if window_function not in {
            "count",
            "mean",
            "sum",
            "max",
            "99%",
            "95%",
            "75%",
            "50%",
            "25%",
            "5%",
            "min",
            "all",
        }:
            raise ValueError(f"window_function value '{window_function}' is not valid")
        return CronLatencySeries(name=name, metric="CRON_LATENCY", window_function=window_function)

    @classmethod
    def stream_message_latency_metric(
        cls,
        name: str,
        window_function: Literal["count", "mean", "sum", "max", "99%", "95%", "75%", "50%", "25%", "5%", "min", "all"],
    ) -> "StreamMessageLatencySeries":
        """Creates a Series of metric kind `StreamMessageLatency`.

        Parameters
        ----------
        name
            A name for your new `stream_message_latency` Series.
        window_function
            The time window to calculate the metric over.

        Returns
        -------
        StreamMessageLatencySeries
            A new `StreamMessageLatencySeries` instance that inherits from the `Series` class.
        """
        if window_function not in {
            "count",
            "mean",
            "sum",
            "max",
            "99%",
            "95%",
            "75%",
            "50%",
            "25%",
            "5%",
            "min",
            "all",
        }:
            raise ValueError(f"window_function value '{window_function}' is not valid")
        return StreamMessageLatencySeries(name=name, metric="STREAM_MESSAGE_LATENCY", window_function=window_function)

    @classmethod
    def stream_messages_processed_metric(
        cls,
        name: str,
        window_function: Literal["count", "mean", "sum", "max", "99%", "95%", "75%", "50%", "25%", "5%", "min", "all"],
    ) -> "StreamMessagesProcessedSeries":
        """Creates a Series of metric kind `StreamMessagesProcessed`.

        Parameters
        ----------
        name
            A name for your new `stream_messages_processed` Series.
        window_function
            The time window to calculate the metric over.

        Returns
        -------
        StreamMessagesProcessedSeries
            A new `StreamMessagesProcessedSeries` instance that inherits from the `Series` class.
        """
        if window_function not in {
            "count",
            "mean",
            "sum",
            "max",
            "99%",
            "95%",
            "75%",
            "50%",
            "25%",
            "5%",
            "min",
            "all",
        }:
            raise ValueError(f"window_function value '{window_function}' is not valid")
        return StreamMessagesProcessedSeries(
            name=name, metric="STREAM_MESSAGES_PROCESSED", window_function=window_function
        )

    @classmethod
    def stream_windows_processed_metric(
        cls,
        name: str,
        window_function: Literal["count", "mean", "sum", "max", "99%", "95%", "75%", "50%", "25%", "5%", "min", "all"],
    ) -> "StreamWindowsProcessedSeries":
        """Creates a Series of metric kind `StreamWindowsProcessed`.

        Parameters
        ----------
        name
            A name for your new `stream_windows_processed` Series.
        window_function
            The time window to calculate the metric over.

        Returns
        -------
        StreamWindowsProcessedSeries
            A new `StreamWindowsProcessedSeries` instance that inherits from the `Series` class.
        """
        if window_function not in {
            "count",
            "mean",
            "sum",
            "max",
            "99%",
            "95%",
            "75%",
            "50%",
            "25%",
            "5%",
            "min",
            "all",
        }:
            raise ValueError(f"window_function value '{window_function}' is not valid")
        return StreamWindowsProcessedSeries(
            name=name, metric="STREAM_WINDOWS_PROCESSED", window_function=window_function
        )

    @classmethod
    def stream_window_latency_metric(
        cls,
        name: str,
        window_function: Literal["count", "mean", "sum", "max", "99%", "95%", "75%", "50%", "25%", "5%", "min", "all"],
    ) -> "StreamWindowLatencySeries":
        """Creates a Series of metric kind `StreamWindowLatency`.

        Parameters
        ----------
        name
            A name for your new `stream_window_latency` Series.
        window_function
            The time window to calculate the metric over.

        Returns
        -------
        StreamWindowLatencySeries
            A new `StreamWindowLatencySeries` instance that inherits from the `Series` class.
        """
        if window_function not in {
            "count",
            "mean",
            "sum",
            "max",
            "99%",
            "95%",
            "75%",
            "50%",
            "25%",
            "5%",
            "min",
            "all",
        }:
            raise ValueError(f"window_function value '{window_function}' is not valid")
        return StreamWindowLatencySeries(name=name, metric="STREAM_WINDOW_LATENCY", window_function=window_function)


class FeatureRequestCountSeries(SeriesBase):
    """
    Series class for metric `feature_request_count`
    """

    def __new__(cls, *args, **kwargs):
        return object.__new__(cls)

    def where(
        self,
        resolver_type: Optional[Union[List[ResolverType], ResolverType]] = None,
        feature_tag: Optional[Union[List[str], str]] = None,
        feature: Optional[Union[List[Any], Any]] = None,
        is_null: Optional[bool] = None,
        feature_status: Optional[Literal["success", "failure"]] = None,
        cache_hit: Optional[bool] = None,
    ) -> "FeatureRequestCountSeries":
        """Attaches a filter to your Series instance.

        Parameters
        ----------
        resolver_type:
            Filters for resolvers by type 'online', 'offline' or 'stream'.
        feature_tag:
            Filters for features matching the given tag(s).
        feature:
            Filters for values pertaining to the given feature.
        is_null:
            Filters for null values.
        feature_status:
            Filters for successes/failures of features.
        cache_hit:
            Filters for cache hits.

        Returns
        -------
        FeatureRequestCountSeries
            A copy of your series with the new filter
        """
        return self._where(
            resolver_type=resolver_type,
            feature_tag=feature_tag,
            feature=feature,
            is_null=is_null,
            feature_status=feature_status,
            cache_hit=cache_hit,
            equals=True,
        )

    def where_not(
        self,
        resolver_type: Optional[Union[List[ResolverType], ResolverType]] = None,
        feature_tag: Optional[Union[List[str], str]] = None,
        feature: Optional[Union[List[Any], Any]] = None,
        is_null: Optional[bool] = None,
        feature_status: Optional[Literal["success", "failure"]] = None,
        cache_hit: Optional[bool] = None,
    ) -> "FeatureRequestCountSeries":
        """Attaches a negative filter to your Series instance.

        Parameters
        ----------
        resolver_type:
            Filters for resolvers by type 'online', 'offline' or 'stream'.
        feature_tag:
            Filters for features matching the given tag(s).
        feature:
            Filters for values pertaining to the given feature.
        is_null:
            Filters for null values.
        feature_status:
            Filters for successes/failures of features.
        cache_hit:
            Filters for cache hits.

        Returns
        -------
        FeatureRequestCountSeries
            A copy of your series with the new filter
        """
        return self._where(
            resolver_type=resolver_type,
            feature_tag=feature_tag,
            feature=feature,
            is_null=is_null,
            feature_status=feature_status,
            cache_hit=cache_hit,
            equals=False,
        )

    def group_by_resolver_type(self) -> "FeatureRequestCountSeries":
        """Attaches a `resolver_type` group-by to your Series instance.

        Returns
        -------
        FeatureRequestCountSeries
            A copy of your series with the new group-by
        """
        copy = self._copy_with()
        copy._group_by.append(GroupByKind.ONLINE_OFFLINE)
        return copy

    def group_by_cache_hit(self) -> "FeatureRequestCountSeries":
        """Attaches a `cache_hit` group-by to your Series instance.

        Returns
        -------
        FeatureRequestCountSeries
            A copy of your series with the new group-by
        """
        copy = self._copy_with()
        copy._group_by.append(GroupByKind.CACHE_HIT)
        return copy

    def group_by_is_null(self) -> "FeatureRequestCountSeries":
        """Attaches a `is_null` group-by to your Series instance.

        Returns
        -------
        FeatureRequestCountSeries
            A copy of your series with the new group-by
        """
        copy = self._copy_with()
        copy._group_by.append(GroupByKind.IS_NULL)
        return copy

    def group_by_feature(self) -> "FeatureRequestCountSeries":
        """Attaches a `feature` group-by to your Series instance.

        Returns
        -------
        FeatureRequestCountSeries
            A copy of your series with the new group-by
        """
        copy = self._copy_with()
        copy._group_by.append(GroupByKind.FEATURE_NAME)
        return copy


class FeatureStalenessSeries(SeriesBase):
    """
    Series class for metric `feature_staleness`
    """

    def __new__(cls, *args, **kwargs):
        return object.__new__(cls)

    def where(
        self,
        resolver_type: Optional[Union[List[ResolverType], ResolverType]] = None,
    ) -> "FeatureStalenessSeries":
        """Attaches a filter to your Series instance.

        Parameters
        ----------
        resolver_type:
            Filters for resolvers by type 'online', 'offline' or 'stream'.

        Returns
        -------
        FeatureStalenessSeries
            A copy of your series with the new filter
        """
        return self._where(
            resolver_type=resolver_type,
            equals=True,
        )

    def where_not(
        self,
        resolver_type: Optional[Union[List[ResolverType], ResolverType]] = None,
    ) -> "FeatureStalenessSeries":
        """Attaches a negative filter to your Series instance.

        Parameters
        ----------
        resolver_type:
            Filters for resolvers by type 'online', 'offline' or 'stream'.

        Returns
        -------
        FeatureStalenessSeries
            A copy of your series with the new filter
        """
        return self._where(
            resolver_type=resolver_type,
            equals=False,
        )

    def group_by_resolver_type(self) -> "FeatureStalenessSeries":
        """Attaches a `resolver_type` group-by to your Series instance.

        Returns
        -------
        FeatureStalenessSeries
            A copy of your series with the new group-by
        """
        copy = self._copy_with()
        copy._group_by.append(GroupByKind.ONLINE_OFFLINE)
        return copy

    def group_by_cache_hit(self) -> "FeatureStalenessSeries":
        """Attaches a `cache_hit` group-by to your Series instance.

        Returns
        -------
        FeatureStalenessSeries
            A copy of your series with the new group-by
        """
        copy = self._copy_with()
        copy._group_by.append(GroupByKind.CACHE_HIT)
        return copy


class FeatureValueSeries(SeriesBase):
    """
    Series class for metric `feature_value`
    """

    def __new__(cls, *args, **kwargs):
        return object.__new__(cls)

    def where(
        self,
        resolver_type: Optional[Union[List[ResolverType], ResolverType]] = None,
        feature_tag: Optional[Union[List[str], str]] = None,
        feature: Optional[Union[List[Any], Any]] = None,
        feature_status: Optional[Literal["success", "failure"]] = None,
        cache_hit: Optional[bool] = None,
    ) -> "FeatureValueSeries":
        """Attaches a filter to your Series instance.

        Parameters
        ----------
        resolver_type:
            Filters for resolvers by type 'online', 'offline' or 'stream'.
        feature_tag:
            Filters for features matching the given tag(s).
        feature:
            Filters for values pertaining to the given feature.
        feature_status:
            Filters for successes/failures of features.
        cache_hit:
            Filters for cache hits.

        Returns
        -------
        FeatureValueSeries
            A copy of your series with the new filter
        """
        return self._where(
            resolver_type=resolver_type,
            feature_tag=feature_tag,
            feature=feature,
            feature_status=feature_status,
            cache_hit=cache_hit,
            equals=True,
        )

    def where_not(
        self,
        resolver_type: Optional[Union[List[ResolverType], ResolverType]] = None,
        feature_tag: Optional[Union[List[str], str]] = None,
        feature: Optional[Union[List[Any], Any]] = None,
        feature_status: Optional[Literal["success", "failure"]] = None,
        cache_hit: Optional[bool] = None,
    ) -> "FeatureValueSeries":
        """Attaches a negative filter to your Series instance.

        Parameters
        ----------
        resolver_type:
            Filters for resolvers by type 'online', 'offline' or 'stream'.
        feature_tag:
            Filters for features matching the given tag(s).
        feature:
            Filters for values pertaining to the given feature.
        feature_status:
            Filters for successes/failures of features.
        cache_hit:
            Filters for cache hits.

        Returns
        -------
        FeatureValueSeries
            A copy of your series with the new filter
        """
        return self._where(
            resolver_type=resolver_type,
            feature_tag=feature_tag,
            feature=feature,
            feature_status=feature_status,
            cache_hit=cache_hit,
            equals=False,
        )

    def group_by_resolver_type(self) -> "FeatureValueSeries":
        """Attaches a `resolver_type` group-by to your Series instance.

        Returns
        -------
        FeatureValueSeries
            A copy of your series with the new group-by
        """
        copy = self._copy_with()
        copy._group_by.append(GroupByKind.ONLINE_OFFLINE)
        return copy

    def group_by_cache_hit(self) -> "FeatureValueSeries":
        """Attaches a `cache_hit` group-by to your Series instance.

        Returns
        -------
        FeatureValueSeries
            A copy of your series with the new group-by
        """
        copy = self._copy_with()
        copy._group_by.append(GroupByKind.CACHE_HIT)
        return copy

    def group_by_feature(self) -> "FeatureValueSeries":
        """Attaches a `feature` group-by to your Series instance.

        Returns
        -------
        FeatureValueSeries
            A copy of your series with the new group-by
        """
        copy = self._copy_with()
        copy._group_by.append(GroupByKind.FEATURE_NAME)
        return copy


class FeatureNullRatioSeries(SeriesBase):
    """
    Series class for metric `feature_null_ratio`
    """

    def __new__(cls, *args, **kwargs):
        return object.__new__(cls)

    def where(
        self,
        resolver_type: Optional[Union[List[ResolverType], ResolverType]] = None,
        feature_tag: Optional[Union[List[str], str]] = None,
        feature: Optional[Union[List[Any], Any]] = None,
        feature_status: Optional[Literal["success", "failure"]] = None,
        cache_hit: Optional[bool] = None,
    ) -> "FeatureNullRatioSeries":
        """Attaches a filter to your Series instance.

        Parameters
        ----------
        resolver_type:
            Filters for resolvers by type 'online', 'offline' or 'stream'.
        feature_tag:
            Filters for features matching the given tag(s).
        feature:
            Filters for values pertaining to the given feature.
        feature_status:
            Filters for successes/failures of features.
        cache_hit:
            Filters for cache hits.

        Returns
        -------
        FeatureNullRatioSeries
            A copy of your series with the new filter
        """
        return self._where(
            resolver_type=resolver_type,
            feature_tag=feature_tag,
            feature=feature,
            feature_status=feature_status,
            cache_hit=cache_hit,
            equals=True,
        )

    def where_not(
        self,
        resolver_type: Optional[Union[List[ResolverType], ResolverType]] = None,
        feature_tag: Optional[Union[List[str], str]] = None,
        feature: Optional[Union[List[Any], Any]] = None,
        feature_status: Optional[Literal["success", "failure"]] = None,
        cache_hit: Optional[bool] = None,
    ) -> "FeatureNullRatioSeries":
        """Attaches a negative filter to your Series instance.

        Parameters
        ----------
        resolver_type:
            Filters for resolvers by type 'online', 'offline' or 'stream'.
        feature_tag:
            Filters for features matching the given tag(s).
        feature:
            Filters for values pertaining to the given feature.
        feature_status:
            Filters for successes/failures of features.
        cache_hit:
            Filters for cache hits.

        Returns
        -------
        FeatureNullRatioSeries
            A copy of your series with the new filter
        """
        return self._where(
            resolver_type=resolver_type,
            feature_tag=feature_tag,
            feature=feature,
            feature_status=feature_status,
            cache_hit=cache_hit,
            equals=False,
        )

    def group_by_resolver_type(self) -> "FeatureNullRatioSeries":
        """Attaches a `resolver_type` group-by to your Series instance.

        Returns
        -------
        FeatureNullRatioSeries
            A copy of your series with the new group-by
        """
        copy = self._copy_with()
        copy._group_by.append(GroupByKind.ONLINE_OFFLINE)
        return copy

    def group_by_cache_hit(self) -> "FeatureNullRatioSeries":
        """Attaches a `cache_hit` group-by to your Series instance.

        Returns
        -------
        FeatureNullRatioSeries
            A copy of your series with the new group-by
        """
        copy = self._copy_with()
        copy._group_by.append(GroupByKind.CACHE_HIT)
        return copy

    def group_by_feature(self) -> "FeatureNullRatioSeries":
        """Attaches a `feature` group-by to your Series instance.

        Returns
        -------
        FeatureNullRatioSeries
            A copy of your series with the new group-by
        """
        copy = self._copy_with()
        copy._group_by.append(GroupByKind.FEATURE_NAME)
        return copy


class ResolverRequestCountSeries(SeriesBase):
    """
    Series class for metric `resolver_request_count`
    """

    def __new__(cls, *args, **kwargs):
        return object.__new__(cls)

    def where(
        self,
        resolver_type: Optional[Union[List[ResolverType], ResolverType]] = None,
        resolver_tag: Optional[Union[List[str], str]] = None,
        resolver: Optional[Union[List[Union[ResolverProtocol, str]], Union[ResolverProtocol, str]]] = None,
        resolver_status: Optional[Literal["success", "failure"]] = None,
    ) -> "ResolverRequestCountSeries":
        """Attaches a filter to your Series instance.

        Parameters
        ----------
        resolver_type:
            Filters for resolvers by type 'online', 'offline' or 'stream'.
        resolver_tag:
            Filters for resolvers matching the given tag(s).
        resolver:
            Filters for values pertaining to the given resolver.
        resolver_status:
            Filters for successes/failures of resolvers.

        Returns
        -------
        ResolverRequestCountSeries
            A copy of your series with the new filter
        """
        return self._where(
            resolver_type=resolver_type,
            resolver_tag=resolver_tag,
            resolver=resolver,
            resolver_status=resolver_status,
            equals=True,
        )

    def where_not(
        self,
        resolver_type: Optional[Union[List[ResolverType], ResolverType]] = None,
        resolver_tag: Optional[Union[List[str], str]] = None,
        resolver: Optional[Union[List[Union[ResolverProtocol, str]], Union[ResolverProtocol, str]]] = None,
        resolver_status: Optional[Literal["success", "failure"]] = None,
    ) -> "ResolverRequestCountSeries":
        """Attaches a negative filter to your Series instance.

        Parameters
        ----------
        resolver_type:
            Filters for resolvers by type 'online', 'offline' or 'stream'.
        resolver_tag:
            Filters for resolvers matching the given tag(s).
        resolver:
            Filters for values pertaining to the given resolver.
        resolver_status:
            Filters for successes/failures of resolvers.

        Returns
        -------
        ResolverRequestCountSeries
            A copy of your series with the new filter
        """
        return self._where(
            resolver_type=resolver_type,
            resolver_tag=resolver_tag,
            resolver=resolver,
            resolver_status=resolver_status,
            equals=False,
        )

    def group_by_resolver_type(self) -> "ResolverRequestCountSeries":
        """Attaches a `resolver_type` group-by to your Series instance.

        Returns
        -------
        ResolverRequestCountSeries
            A copy of your series with the new group-by
        """
        copy = self._copy_with()
        copy._group_by.append(GroupByKind.ONLINE_OFFLINE)
        return copy

    def group_by_cache_hit(self) -> "ResolverRequestCountSeries":
        """Attaches a `cache_hit` group-by to your Series instance.

        Returns
        -------
        ResolverRequestCountSeries
            A copy of your series with the new group-by
        """
        copy = self._copy_with()
        copy._group_by.append(GroupByKind.CACHE_HIT)
        return copy

    def group_by_resolver(self) -> "ResolverRequestCountSeries":
        """Attaches a `resolver` group-by to your Series instance.

        Returns
        -------
        ResolverRequestCountSeries
            A copy of your series with the new group-by
        """
        copy = self._copy_with()
        copy._group_by.append(GroupByKind.RESOLVER_NAME)
        return copy

    def group_by_resolver_status(self) -> "ResolverRequestCountSeries":
        """Attaches a `resolver_status` group-by to your Series instance.

        Returns
        -------
        ResolverRequestCountSeries
            A copy of your series with the new group-by
        """
        copy = self._copy_with()
        copy._group_by.append(GroupByKind.RESOLVER_STATUS)
        return copy


class ResolverLatencySeries(SeriesBase):
    """
    Series class for metric `resolver_latency`
    """

    def __new__(cls, *args, **kwargs):
        return object.__new__(cls)

    def where(
        self,
        resolver_type: Optional[Union[List[ResolverType], ResolverType]] = None,
        resolver_tag: Optional[Union[List[str], str]] = None,
        resolver: Optional[Union[List[Union[ResolverProtocol, str]], Union[ResolverProtocol, str]]] = None,
        resolver_status: Optional[Literal["success", "failure"]] = None,
    ) -> "ResolverLatencySeries":
        """Attaches a filter to your Series instance.

        Parameters
        ----------
        resolver_type:
            Filters for resolvers by type 'online', 'offline' or 'stream'.
        resolver_tag:
            Filters for resolvers matching the given tag(s).
        resolver:
            Filters for values pertaining to the given resolver.
        resolver_status:
            Filters for successes/failures of resolvers.

        Returns
        -------
        ResolverLatencySeries
            A copy of your series with the new filter
        """
        return self._where(
            resolver_type=resolver_type,
            resolver_tag=resolver_tag,
            resolver=resolver,
            resolver_status=resolver_status,
            equals=True,
        )

    def where_not(
        self,
        resolver_type: Optional[Union[List[ResolverType], ResolverType]] = None,
        resolver_tag: Optional[Union[List[str], str]] = None,
        resolver: Optional[Union[List[Union[ResolverProtocol, str]], Union[ResolverProtocol, str]]] = None,
        resolver_status: Optional[Literal["success", "failure"]] = None,
    ) -> "ResolverLatencySeries":
        """Attaches a negative filter to your Series instance.

        Parameters
        ----------
        resolver_type:
            Filters for resolvers by type 'online', 'offline' or 'stream'.
        resolver_tag:
            Filters for resolvers matching the given tag(s).
        resolver:
            Filters for values pertaining to the given resolver.
        resolver_status:
            Filters for successes/failures of resolvers.

        Returns
        -------
        ResolverLatencySeries
            A copy of your series with the new filter
        """
        return self._where(
            resolver_type=resolver_type,
            resolver_tag=resolver_tag,
            resolver=resolver,
            resolver_status=resolver_status,
            equals=False,
        )

    def group_by_resolver_type(self) -> "ResolverLatencySeries":
        """Attaches a `resolver_type` group-by to your Series instance.

        Returns
        -------
        ResolverLatencySeries
            A copy of your series with the new group-by
        """
        copy = self._copy_with()
        copy._group_by.append(GroupByKind.ONLINE_OFFLINE)
        return copy

    def group_by_cache_hit(self) -> "ResolverLatencySeries":
        """Attaches a `cache_hit` group-by to your Series instance.

        Returns
        -------
        ResolverLatencySeries
            A copy of your series with the new group-by
        """
        copy = self._copy_with()
        copy._group_by.append(GroupByKind.CACHE_HIT)
        return copy

    def group_by_resolver(self) -> "ResolverLatencySeries":
        """Attaches a `resolver` group-by to your Series instance.

        Returns
        -------
        ResolverLatencySeries
            A copy of your series with the new group-by
        """
        copy = self._copy_with()
        copy._group_by.append(GroupByKind.RESOLVER_NAME)
        return copy

    def group_by_resolver_status(self) -> "ResolverLatencySeries":
        """Attaches a `resolver_status` group-by to your Series instance.

        Returns
        -------
        ResolverLatencySeries
            A copy of your series with the new group-by
        """
        copy = self._copy_with()
        copy._group_by.append(GroupByKind.RESOLVER_STATUS)
        return copy


class ResolverSuccessRatioSeries(SeriesBase):
    """
    Series class for metric `resolver_success_ratio`
    """

    def __new__(cls, *args, **kwargs):
        return object.__new__(cls)

    def where(
        self,
        resolver_type: Optional[Union[List[ResolverType], ResolverType]] = None,
        resolver_tag: Optional[Union[List[str], str]] = None,
        resolver: Optional[Union[List[Union[ResolverProtocol, str]], Union[ResolverProtocol, str]]] = None,
        resolver_status: Optional[Literal["success", "failure"]] = None,
    ) -> "ResolverSuccessRatioSeries":
        """Attaches a filter to your Series instance.

        Parameters
        ----------
        resolver_type:
            Filters for resolvers by type 'online', 'offline' or 'stream'.
        resolver_tag:
            Filters for resolvers matching the given tag(s).
        resolver:
            Filters for values pertaining to the given resolver.
        resolver_status:
            Filters for successes/failures of resolvers.

        Returns
        -------
        ResolverSuccessRatioSeries
            A copy of your series with the new filter
        """
        return self._where(
            resolver_type=resolver_type,
            resolver_tag=resolver_tag,
            resolver=resolver,
            resolver_status=resolver_status,
            equals=True,
        )

    def where_not(
        self,
        resolver_type: Optional[Union[List[ResolverType], ResolverType]] = None,
        resolver_tag: Optional[Union[List[str], str]] = None,
        resolver: Optional[Union[List[Union[ResolverProtocol, str]], Union[ResolverProtocol, str]]] = None,
        resolver_status: Optional[Literal["success", "failure"]] = None,
    ) -> "ResolverSuccessRatioSeries":
        """Attaches a negative filter to your Series instance.

        Parameters
        ----------
        resolver_type:
            Filters for resolvers by type 'online', 'offline' or 'stream'.
        resolver_tag:
            Filters for resolvers matching the given tag(s).
        resolver:
            Filters for values pertaining to the given resolver.
        resolver_status:
            Filters for successes/failures of resolvers.

        Returns
        -------
        ResolverSuccessRatioSeries
            A copy of your series with the new filter
        """
        return self._where(
            resolver_type=resolver_type,
            resolver_tag=resolver_tag,
            resolver=resolver,
            resolver_status=resolver_status,
            equals=False,
        )

    def group_by_resolver_type(self) -> "ResolverSuccessRatioSeries":
        """Attaches a `resolver_type` group-by to your Series instance.

        Returns
        -------
        ResolverSuccessRatioSeries
            A copy of your series with the new group-by
        """
        copy = self._copy_with()
        copy._group_by.append(GroupByKind.ONLINE_OFFLINE)
        return copy

    def group_by_cache_hit(self) -> "ResolverSuccessRatioSeries":
        """Attaches a `cache_hit` group-by to your Series instance.

        Returns
        -------
        ResolverSuccessRatioSeries
            A copy of your series with the new group-by
        """
        copy = self._copy_with()
        copy._group_by.append(GroupByKind.CACHE_HIT)
        return copy

    def group_by_resolver(self) -> "ResolverSuccessRatioSeries":
        """Attaches a `resolver` group-by to your Series instance.

        Returns
        -------
        ResolverSuccessRatioSeries
            A copy of your series with the new group-by
        """
        copy = self._copy_with()
        copy._group_by.append(GroupByKind.RESOLVER_NAME)
        return copy

    def group_by_resolver_status(self) -> "ResolverSuccessRatioSeries":
        """Attaches a `resolver_status` group-by to your Series instance.

        Returns
        -------
        ResolverSuccessRatioSeries
            A copy of your series with the new group-by
        """
        copy = self._copy_with()
        copy._group_by.append(GroupByKind.RESOLVER_STATUS)
        return copy


class QueryCountSeries(SeriesBase):
    """
    Series class for metric `query_count`
    """

    def __new__(cls, *args, **kwargs):
        return object.__new__(cls)

    def where(
        self,
        resolver_type: Optional[Union[List[ResolverType], ResolverType]] = None,
        query_name: Optional[Union[List[str], str]] = None,
        query_status: Optional[Literal["success", "failure"]] = None,
    ) -> "QueryCountSeries":
        """Attaches a filter to your Series instance.

        Parameters
        ----------
        resolver_type:
            Filters for resolvers by type 'online', 'offline' or 'stream'.
        query_name:
            Filters for queries matching the given name(s).
        query_status:
            Filters for successes/failures of queries.

        Returns
        -------
        QueryCountSeries
            A copy of your series with the new filter
        """
        return self._where(
            resolver_type=resolver_type,
            query_name=query_name,
            query_status=query_status,
            equals=True,
        )

    def where_not(
        self,
        resolver_type: Optional[Union[List[ResolverType], ResolverType]] = None,
        query_name: Optional[Union[List[str], str]] = None,
        query_status: Optional[Literal["success", "failure"]] = None,
    ) -> "QueryCountSeries":
        """Attaches a negative filter to your Series instance.

        Parameters
        ----------
        resolver_type:
            Filters for resolvers by type 'online', 'offline' or 'stream'.
        query_name:
            Filters for queries matching the given name(s).
        query_status:
            Filters for successes/failures of queries.

        Returns
        -------
        QueryCountSeries
            A copy of your series with the new filter
        """
        return self._where(
            resolver_type=resolver_type,
            query_name=query_name,
            query_status=query_status,
            equals=False,
        )

    def group_by_query_status(self) -> "QueryCountSeries":
        """Attaches a `query_status` group-by to your Series instance.

        Returns
        -------
        QueryCountSeries
            A copy of your series with the new group-by
        """
        copy = self._copy_with()
        copy._group_by.append(GroupByKind.QUERY_STATUS)
        return copy

    def group_by_query_name(self) -> "QueryCountSeries":
        """Attaches a `query_name` group-by to your Series instance.

        Returns
        -------
        QueryCountSeries
            A copy of your series with the new group-by
        """
        copy = self._copy_with()
        copy._group_by.append(GroupByKind.QUERY_NAME)
        return copy


class QueryLatencySeries(SeriesBase):
    """
    Series class for metric `query_latency`
    """

    def __new__(cls, *args, **kwargs):
        return object.__new__(cls)

    def where(
        self,
        resolver_type: Optional[Union[List[ResolverType], ResolverType]] = None,
        query_name: Optional[Union[List[str], str]] = None,
        query_status: Optional[Literal["success", "failure"]] = None,
    ) -> "QueryLatencySeries":
        """Attaches a filter to your Series instance.

        Parameters
        ----------
        resolver_type:
            Filters for resolvers by type 'online', 'offline' or 'stream'.
        query_name:
            Filters for queries matching the given name(s).
        query_status:
            Filters for successes/failures of queries.

        Returns
        -------
        QueryLatencySeries
            A copy of your series with the new filter
        """
        return self._where(
            resolver_type=resolver_type,
            query_name=query_name,
            query_status=query_status,
            equals=True,
        )

    def where_not(
        self,
        resolver_type: Optional[Union[List[ResolverType], ResolverType]] = None,
        query_name: Optional[Union[List[str], str]] = None,
        query_status: Optional[Literal["success", "failure"]] = None,
    ) -> "QueryLatencySeries":
        """Attaches a negative filter to your Series instance.

        Parameters
        ----------
        resolver_type:
            Filters for resolvers by type 'online', 'offline' or 'stream'.
        query_name:
            Filters for queries matching the given name(s).
        query_status:
            Filters for successes/failures of queries.

        Returns
        -------
        QueryLatencySeries
            A copy of your series with the new filter
        """
        return self._where(
            resolver_type=resolver_type,
            query_name=query_name,
            query_status=query_status,
            equals=False,
        )

    def group_by_query_status(self) -> "QueryLatencySeries":
        """Attaches a `query_status` group-by to your Series instance.

        Returns
        -------
        QueryLatencySeries
            A copy of your series with the new group-by
        """
        copy = self._copy_with()
        copy._group_by.append(GroupByKind.QUERY_STATUS)
        return copy

    def group_by_query_name(self) -> "QueryLatencySeries":
        """Attaches a `query_name` group-by to your Series instance.

        Returns
        -------
        QueryLatencySeries
            A copy of your series with the new group-by
        """
        copy = self._copy_with()
        copy._group_by.append(GroupByKind.QUERY_NAME)
        return copy


class QuerySuccessRatioSeries(SeriesBase):
    """
    Series class for metric `query_success_ratio`
    """

    def __new__(cls, *args, **kwargs):
        return object.__new__(cls)

    def where(
        self,
        resolver_type: Optional[Union[List[ResolverType], ResolverType]] = None,
        query_name: Optional[Union[List[str], str]] = None,
    ) -> "QuerySuccessRatioSeries":
        """Attaches a filter to your Series instance.

        Parameters
        ----------
        resolver_type:
            Filters for resolvers by type 'online', 'offline' or 'stream'.
        query_name:
            Filters for queries matching the given name(s).

        Returns
        -------
        QuerySuccessRatioSeries
            A copy of your series with the new filter
        """
        return self._where(
            resolver_type=resolver_type,
            query_name=query_name,
            equals=True,
        )

    def where_not(
        self,
        resolver_type: Optional[Union[List[ResolverType], ResolverType]] = None,
        query_name: Optional[Union[List[str], str]] = None,
    ) -> "QuerySuccessRatioSeries":
        """Attaches a negative filter to your Series instance.

        Parameters
        ----------
        resolver_type:
            Filters for resolvers by type 'online', 'offline' or 'stream'.
        query_name:
            Filters for queries matching the given name(s).

        Returns
        -------
        QuerySuccessRatioSeries
            A copy of your series with the new filter
        """
        return self._where(
            resolver_type=resolver_type,
            query_name=query_name,
            equals=False,
        )

    def group_by_query_name(self) -> "QuerySuccessRatioSeries":
        """Attaches a `query_name` group-by to your Series instance.

        Returns
        -------
        QuerySuccessRatioSeries
            A copy of your series with the new group-by
        """
        copy = self._copy_with()
        copy._group_by.append(GroupByKind.QUERY_NAME)
        return copy


class CronCountSeries(SeriesBase):
    """
    Series class for metric `cron_count`
    """

    def __new__(cls, *args, **kwargs):
        return object.__new__(cls)

    def where(
        self,
        resolver_type: Optional[Union[List[ResolverType], ResolverType]] = None,
        resolver_tag: Optional[Union[List[str], str]] = None,
        resolver: Optional[Union[List[Union[ResolverProtocol, str]], Union[ResolverProtocol, str]]] = None,
        resolver_status: Optional[Literal["success", "failure"]] = None,
    ) -> "CronCountSeries":
        """Attaches a filter to your Series instance.

        Parameters
        ----------
        resolver_type:
            Filters for resolvers by type 'online', 'offline' or 'stream'.
        resolver_tag:
            Filters for resolvers matching the given tag(s).
        resolver:
            Filters for values pertaining to the given resolver.
        resolver_status:
            Filters for successes/failures of resolvers.

        Returns
        -------
        CronCountSeries
            A copy of your series with the new filter
        """
        return self._where(
            resolver_type=resolver_type,
            resolver_tag=resolver_tag,
            resolver=resolver,
            resolver_status=resolver_status,
            equals=True,
        )

    def where_not(
        self,
        resolver_type: Optional[Union[List[ResolverType], ResolverType]] = None,
        resolver_tag: Optional[Union[List[str], str]] = None,
        resolver: Optional[Union[List[Union[ResolverProtocol, str]], Union[ResolverProtocol, str]]] = None,
        resolver_status: Optional[Literal["success", "failure"]] = None,
    ) -> "CronCountSeries":
        """Attaches a negative filter to your Series instance.

        Parameters
        ----------
        resolver_type:
            Filters for resolvers by type 'online', 'offline' or 'stream'.
        resolver_tag:
            Filters for resolvers matching the given tag(s).
        resolver:
            Filters for values pertaining to the given resolver.
        resolver_status:
            Filters for successes/failures of resolvers.

        Returns
        -------
        CronCountSeries
            A copy of your series with the new filter
        """
        return self._where(
            resolver_type=resolver_type,
            resolver_tag=resolver_tag,
            resolver=resolver,
            resolver_status=resolver_status,
            equals=False,
        )


class CronLatencySeries(SeriesBase):
    """
    Series class for metric `cron_latency`
    """

    def __new__(cls, *args, **kwargs):
        return object.__new__(cls)

    def where(
        self,
        resolver_type: Optional[Union[List[ResolverType], ResolverType]] = None,
        resolver_tag: Optional[Union[List[str], str]] = None,
        resolver: Optional[Union[List[Union[ResolverProtocol, str]], Union[ResolverProtocol, str]]] = None,
        resolver_status: Optional[Literal["success", "failure"]] = None,
    ) -> "CronLatencySeries":
        """Attaches a filter to your Series instance.

        Parameters
        ----------
        resolver_type:
            Filters for resolvers by type 'online', 'offline' or 'stream'.
        resolver_tag:
            Filters for resolvers matching the given tag(s).
        resolver:
            Filters for values pertaining to the given resolver.
        resolver_status:
            Filters for successes/failures of resolvers.

        Returns
        -------
        CronLatencySeries
            A copy of your series with the new filter
        """
        return self._where(
            resolver_type=resolver_type,
            resolver_tag=resolver_tag,
            resolver=resolver,
            resolver_status=resolver_status,
            equals=True,
        )

    def where_not(
        self,
        resolver_type: Optional[Union[List[ResolverType], ResolverType]] = None,
        resolver_tag: Optional[Union[List[str], str]] = None,
        resolver: Optional[Union[List[Union[ResolverProtocol, str]], Union[ResolverProtocol, str]]] = None,
        resolver_status: Optional[Literal["success", "failure"]] = None,
    ) -> "CronLatencySeries":
        """Attaches a negative filter to your Series instance.

        Parameters
        ----------
        resolver_type:
            Filters for resolvers by type 'online', 'offline' or 'stream'.
        resolver_tag:
            Filters for resolvers matching the given tag(s).
        resolver:
            Filters for values pertaining to the given resolver.
        resolver_status:
            Filters for successes/failures of resolvers.

        Returns
        -------
        CronLatencySeries
            A copy of your series with the new filter
        """
        return self._where(
            resolver_type=resolver_type,
            resolver_tag=resolver_tag,
            resolver=resolver,
            resolver_status=resolver_status,
            equals=False,
        )

    def group_by_resolver_type(self) -> "CronLatencySeries":
        """Attaches a `resolver_type` group-by to your Series instance.

        Returns
        -------
        CronLatencySeries
            A copy of your series with the new group-by
        """
        copy = self._copy_with()
        copy._group_by.append(GroupByKind.ONLINE_OFFLINE)
        return copy

    def group_by_cache_hit(self) -> "CronLatencySeries":
        """Attaches a `cache_hit` group-by to your Series instance.

        Returns
        -------
        CronLatencySeries
            A copy of your series with the new group-by
        """
        copy = self._copy_with()
        copy._group_by.append(GroupByKind.CACHE_HIT)
        return copy


class StreamMessageLatencySeries(SeriesBase):
    """
    Series class for metric `stream_message_latency`
    """

    def __new__(cls, *args, **kwargs):
        return object.__new__(cls)

    def where(
        self,
        resolver_tag: Optional[Union[List[str], str]] = None,
        resolver: Optional[Union[List[Union[ResolverProtocol, str]], Union[ResolverProtocol, str]]] = None,
        resolver_status: Optional[Literal["success", "failure"]] = None,
    ) -> "StreamMessageLatencySeries":
        """Attaches a filter to your Series instance.

        Parameters
        ----------
        resolver_tag:
            Filters for resolvers matching the given tag(s).
        resolver:
            Filters for values pertaining to the given resolver.
        resolver_status:
            Filters for successes/failures of resolvers.

        Returns
        -------
        StreamMessageLatencySeries
            A copy of your series with the new filter
        """
        return self._where(
            resolver_tag=resolver_tag,
            resolver=resolver,
            resolver_status=resolver_status,
            equals=True,
        )

    def where_not(
        self,
        resolver_tag: Optional[Union[List[str], str]] = None,
        resolver: Optional[Union[List[Union[ResolverProtocol, str]], Union[ResolverProtocol, str]]] = None,
        resolver_status: Optional[Literal["success", "failure"]] = None,
    ) -> "StreamMessageLatencySeries":
        """Attaches a negative filter to your Series instance.

        Parameters
        ----------
        resolver_tag:
            Filters for resolvers matching the given tag(s).
        resolver:
            Filters for values pertaining to the given resolver.
        resolver_status:
            Filters for successes/failures of resolvers.

        Returns
        -------
        StreamMessageLatencySeries
            A copy of your series with the new filter
        """
        return self._where(
            resolver_tag=resolver_tag,
            resolver=resolver,
            resolver_status=resolver_status,
            equals=False,
        )

    def group_by_resolver_status(self) -> "StreamMessageLatencySeries":
        """Attaches a `resolver_status` group-by to your Series instance.

        Returns
        -------
        StreamMessageLatencySeries
            A copy of your series with the new group-by
        """
        copy = self._copy_with()
        copy._group_by.append(GroupByKind.RESOLVER_STATUS)
        return copy


class StreamMessagesProcessedSeries(SeriesBase):
    """
    Series class for metric `stream_messages_processed`
    """

    def __new__(cls, *args, **kwargs):
        return object.__new__(cls)

    def where(
        self,
        resolver_tag: Optional[Union[List[str], str]] = None,
        resolver: Optional[Union[List[Union[ResolverProtocol, str]], Union[ResolverProtocol, str]]] = None,
        resolver_status: Optional[Literal["success", "failure"]] = None,
    ) -> "StreamMessagesProcessedSeries":
        """Attaches a filter to your Series instance.

        Parameters
        ----------
        resolver_tag:
            Filters for resolvers matching the given tag(s).
        resolver:
            Filters for values pertaining to the given resolver.
        resolver_status:
            Filters for successes/failures of resolvers.

        Returns
        -------
        StreamMessagesProcessedSeries
            A copy of your series with the new filter
        """
        return self._where(
            resolver_tag=resolver_tag,
            resolver=resolver,
            resolver_status=resolver_status,
            equals=True,
        )

    def where_not(
        self,
        resolver_tag: Optional[Union[List[str], str]] = None,
        resolver: Optional[Union[List[Union[ResolverProtocol, str]], Union[ResolverProtocol, str]]] = None,
        resolver_status: Optional[Literal["success", "failure"]] = None,
    ) -> "StreamMessagesProcessedSeries":
        """Attaches a negative filter to your Series instance.

        Parameters
        ----------
        resolver_tag:
            Filters for resolvers matching the given tag(s).
        resolver:
            Filters for values pertaining to the given resolver.
        resolver_status:
            Filters for successes/failures of resolvers.

        Returns
        -------
        StreamMessagesProcessedSeries
            A copy of your series with the new filter
        """
        return self._where(
            resolver_tag=resolver_tag,
            resolver=resolver,
            resolver_status=resolver_status,
            equals=False,
        )

    def group_by_resolver_status(self) -> "StreamMessagesProcessedSeries":
        """Attaches a `resolver_status` group-by to your Series instance.

        Returns
        -------
        StreamMessagesProcessedSeries
            A copy of your series with the new group-by
        """
        copy = self._copy_with()
        copy._group_by.append(GroupByKind.RESOLVER_STATUS)
        return copy


class StreamWindowsProcessedSeries(SeriesBase):
    """
    Series class for metric `stream_windows_processed`
    """

    def __new__(cls, *args, **kwargs):
        return object.__new__(cls)

    def where(
        self,
        resolver_tag: Optional[Union[List[str], str]] = None,
        resolver: Optional[Union[List[Union[ResolverProtocol, str]], Union[ResolverProtocol, str]]] = None,
        resolver_status: Optional[Literal["success", "failure"]] = None,
    ) -> "StreamWindowsProcessedSeries":
        """Attaches a filter to your Series instance.

        Parameters
        ----------
        resolver_tag:
            Filters for resolvers matching the given tag(s).
        resolver:
            Filters for values pertaining to the given resolver.
        resolver_status:
            Filters for successes/failures of resolvers.

        Returns
        -------
        StreamWindowsProcessedSeries
            A copy of your series with the new filter
        """
        return self._where(
            resolver_tag=resolver_tag,
            resolver=resolver,
            resolver_status=resolver_status,
            equals=True,
        )

    def where_not(
        self,
        resolver_tag: Optional[Union[List[str], str]] = None,
        resolver: Optional[Union[List[Union[ResolverProtocol, str]], Union[ResolverProtocol, str]]] = None,
        resolver_status: Optional[Literal["success", "failure"]] = None,
    ) -> "StreamWindowsProcessedSeries":
        """Attaches a negative filter to your Series instance.

        Parameters
        ----------
        resolver_tag:
            Filters for resolvers matching the given tag(s).
        resolver:
            Filters for values pertaining to the given resolver.
        resolver_status:
            Filters for successes/failures of resolvers.

        Returns
        -------
        StreamWindowsProcessedSeries
            A copy of your series with the new filter
        """
        return self._where(
            resolver_tag=resolver_tag,
            resolver=resolver,
            resolver_status=resolver_status,
            equals=False,
        )

    def group_by_resolver_status(self) -> "StreamWindowsProcessedSeries":
        """Attaches a `resolver_status` group-by to your Series instance.

        Returns
        -------
        StreamWindowsProcessedSeries
            A copy of your series with the new group-by
        """
        copy = self._copy_with()
        copy._group_by.append(GroupByKind.RESOLVER_STATUS)
        return copy


class StreamWindowLatencySeries(SeriesBase):
    """
    Series class for metric `stream_window_latency`
    """

    def __new__(cls, *args, **kwargs):
        return object.__new__(cls)

    def where(
        self,
        resolver_tag: Optional[Union[List[str], str]] = None,
        resolver: Optional[Union[List[Union[ResolverProtocol, str]], Union[ResolverProtocol, str]]] = None,
        resolver_status: Optional[Literal["success", "failure"]] = None,
    ) -> "StreamWindowLatencySeries":
        """Attaches a filter to your Series instance.

        Parameters
        ----------
        resolver_tag:
            Filters for resolvers matching the given tag(s).
        resolver:
            Filters for values pertaining to the given resolver.
        resolver_status:
            Filters for successes/failures of resolvers.

        Returns
        -------
        StreamWindowLatencySeries
            A copy of your series with the new filter
        """
        return self._where(
            resolver_tag=resolver_tag,
            resolver=resolver,
            resolver_status=resolver_status,
            equals=True,
        )

    def where_not(
        self,
        resolver_tag: Optional[Union[List[str], str]] = None,
        resolver: Optional[Union[List[Union[ResolverProtocol, str]], Union[ResolverProtocol, str]]] = None,
        resolver_status: Optional[Literal["success", "failure"]] = None,
    ) -> "StreamWindowLatencySeries":
        """Attaches a negative filter to your Series instance.

        Parameters
        ----------
        resolver_tag:
            Filters for resolvers matching the given tag(s).
        resolver:
            Filters for values pertaining to the given resolver.
        resolver_status:
            Filters for successes/failures of resolvers.

        Returns
        -------
        StreamWindowLatencySeries
            A copy of your series with the new filter
        """
        return self._where(
            resolver_tag=resolver_tag,
            resolver=resolver,
            resolver_status=resolver_status,
            equals=False,
        )

    def group_by_resolver_status(self) -> "StreamWindowLatencySeries":
        """Attaches a `resolver_status` group-by to your Series instance.

        Returns
        -------
        StreamWindowLatencySeries
            A copy of your series with the new group-by
        """
        copy = self._copy_with()
        copy._group_by.append(GroupByKind.RESOLVER_STATUS)
        return copy



import dataclasses
import zlib
from copy import deepcopy
from typing import Any, Callable, ClassVar, List, Literal, Optional, Set, Tuple, TypeVar, Union, overload

from typing_extensions import ParamSpec

from chalk._monitoring.charts_codegen import Series
from chalk._monitoring.charts_enums_codegen import (
    AlertSeverityKind,
    ChartLinkKind,
    MetricFormulaKind,
    ThresholdPosition,
)
from chalk._monitoring.charts_series_base import SeriesBase, ThresholdFunction
from chalk.features import FeatureWrapper
from chalk.features.resolver import Resolver, ResolverProtocol
from chalk.utils.duration import parse_chalk_duration


@dataclasses.dataclass
class _SingleSeriesOperand:
    operand: int


@dataclasses.dataclass
class _MultiSeriesOperand:
    operands: List[int]


@dataclasses.dataclass
class _DatasetFeatureOperand:
    dataset: str
    feature: str


class _Formula:
    def __init__(
        self,
        name: Optional[str] = None,
        kind: Optional[Union[MetricFormulaKind, str]] = None,
        operands: Optional[Union[_SingleSeriesOperand, _MultiSeriesOperand, _DatasetFeatureOperand]] = None,
    ):
        self._name = name
        self._kind = MetricFormulaKind(kind.upper()) if kind else None
        self._operands = operands

    def with_name(self, name: str) -> "_Formula":
        copy = self._copy_with()
        copy._name = name
        return copy

    def with_kind(self, kind: Union[MetricFormulaKind, str]) -> "_Formula":
        copy = self._copy_with()
        copy._kind = MetricFormulaKind(kind.upper())
        return copy

    def with_operands(
        self, operands: Union[_SingleSeriesOperand, _MultiSeriesOperand, _DatasetFeatureOperand]
    ) -> "_Formula":
        copy = self._copy_with()
        copy._operands = operands
        return copy

    def _copy_with(self) -> "_Formula":
        self_copy = deepcopy(self)
        return self_copy

    def __hash__(self) -> int:
        name = self._name if self._name else "."
        kind = str(self._kind) if self._kind else "."
        operands = ""
        if isinstance(self._operands, _SingleSeriesOperand):
            operands = self._operands.operand
        elif isinstance(self._operands, _MultiSeriesOperand):
            operands = self._operands.operands
        elif isinstance(self._operands, _DatasetFeatureOperand):
            operands = f"{self._operands.dataset}.{self._operands.feature}"

        formula_string = f"formula.{name}.{kind}.{operands}"

        return zlib.crc32(formula_string.encode())


class Trigger:
    """
    Class to attach an alert to a Chart. Usually instantiated with the `Chart(...).with_trigger(Trigger(...))` method.
    """

    def __init__(
        self,
        name: str,
        severity: Optional[
            Literal[
                "CRITICAL",
                "ERROR",
                "WARNING",
                "INFO",
            ]
        ] = None,
        threshold_position: Optional[Union[ThresholdPosition, str]] = None,
        threshold_value: Optional[float] = None,
        series_name: Optional[str] = None,
        channel_name: Optional[str] = None,
    ):
        self._name = name
        self._severity = severity and AlertSeverityKind(severity.upper())
        self._threshold_position = threshold_position
        self._threshold_value = threshold_value
        self._series_name = series_name
        self._channel_name = channel_name

    def with_name(self, name: str) -> "Trigger":
        copy = self._copy_with()
        copy._name = name
        return copy

    def with_severity(
        self,
        severity: Literal[
            "CRITICAL",
            "ERROR",
            "WARNING",
            "INFO",
        ],
    ) -> "Trigger":
        copy = self._copy_with()
        copy._severity = AlertSeverityKind(severity.upper())
        return copy

    def with_critical_severity(self) -> "Trigger":
        copy = self._copy_with()
        copy._severity = AlertSeverityKind.CRITICAL
        return copy

    def with_error_severity(self) -> "Trigger":
        copy = self._copy_with()
        copy._severity = AlertSeverityKind.ERROR
        return copy

    def with_warning_severity(self) -> "Trigger":
        copy = self._copy_with()
        copy._severity = AlertSeverityKind.WARNING
        return copy

    def with_info_severity(self) -> "Trigger":
        copy = self._copy_with()
        copy._severity = AlertSeverityKind.INFO
        return copy

    def with_threshold_position(self, threshold_position: Union[ThresholdPosition, str]) -> "Trigger":
        copy = self._copy_with()
        copy._threshold_position = ThresholdPosition(threshold_position.upper())
        return copy

    def with_threshold_value(self, threshold_value: float) -> "Trigger":
        copy = self._copy_with()
        copy._threshold_value = threshold_value
        return copy

    def with_series_name(self, series_name: str) -> "Trigger":
        copy = self._copy_with()
        copy._series_name = series_name
        return copy

    def with_channel_name(self, channel_name: str) -> "Trigger":
        copy = self._copy_with()
        copy._channel_name = channel_name
        return copy

    def _copy_with(self) -> "Trigger":
        self_copy = deepcopy(self)
        return self_copy

    def __str__(self) -> str:
        return f"Trigger(name='{self._name}')"

    def __hash__(self) -> int:
        name = self._name if self._name else "."
        severity = str(self._severity) if self._severity else "."
        threshold_position = str(self._threshold_position) if self._threshold_position else "."
        threshold_value = str(self._threshold_value) if self._threshold_value else "."
        series_name = self._series_name if self._series_name else "."
        channel_name = self._channel_name if self._channel_name else "."

        trigger_string = (
            f"trigger.{name}.{severity}.{threshold_position}." f"{threshold_value}.{series_name}.{channel_name}"
        )

        return zlib.crc32(trigger_string.encode())


P = ParamSpec("P")
T = TypeVar("T")


def _copy_with(function: Callable[P, "Chart"]) -> Callable[P, "Chart"]:
    def inner(self, *args, **kwargs):
        copy = deepcopy(self)
        if not self._keep:
            if self in Chart._registry:
                Chart._registry.remove(self)
        return_copy = function(copy, *args, **kwargs)
        Chart._registry.add(return_copy)
        return return_copy

    return inner


# MetricConfigGQL
class Chart:
    """
    Class describing a single visual metric.
    """

    _registry: ClassVar[Set[Union[str, "Chart"]]] = set()

    def __init__(self, name: str, window_period: str = "1h", keep: Optional[bool] = False):
        """

        Parameters
        ----------
        name
            The name of the chart
        window_period
            The length of the window, e.g. "20m" or "1h"
        keep
            By default, the builder methods that return Charts make a deepcopy
            and be deregistered from deployment. If `keep=True`, this chart
            and all descendant charts will be registered automatically.

        Returns
        -------
        Chart
            Your first instance of a Chart!
        """
        self._name = name
        self._window_period = window_period
        self._series: List[SeriesBase] = []
        self._formulas: List[_Formula] = []
        self._trigger = None
        self._keep = keep
        self._entity_id = None
        self._entity_kind = ChartLinkKind.manual
        Chart._registry.add(self)

    @_copy_with
    def with_name(self, name: str) -> "Chart":
        """Override the name of a chart.

        Parameters
        ----------
        name
            A new name for a chart.

        Returns
        -------
        Chart
            A copy of your chart with the new name
        """
        self._name = name
        return self

    @_copy_with
    def with_window_period(self, window_period: str) -> "Chart":
        """Change the window period for a chart.

        Parameters
        ----------
        window_period
            A new window period for a chart, e.g. "20m" or "1h".

        Returns
        -------
        Chart
            A copy of your chart with the new window period
        """
        parse_chalk_duration(window_period)
        self._window_period = window_period
        return self

    @_copy_with
    def with_series(self, series: SeriesBase) -> "Chart":
        """Attaches a `Series` to your `Chart` instance.

        Parameters
        ----------
        series
            A Series instance to attach to the Chart.
            A Chart can have any number of Series.

        Returns
        -------
        Chart
            A copy of your chart with the new name
        """
        if not isinstance(series, SeriesBase):
            raise ValueError(f"'series' value '{series}' must be a Series object")
        self._series.append(series)
        if series._entity_id:
            self._entity_id = series._entity_id
            self._entity_kind = series._entity_kind
        return self

    def get_series(self, series_name: str) -> SeriesBase:
        """Get a `Series` from your `Chart` by series name.

        It is advised to use different series names within your charts.

        Parameters
        ----------
        series_name
            The name of the Series

        Returns
        -------
        Series
            The first series added to your `Chart` with the given series name.
        """
        for series in self._series:
            if series._name == series_name:
                return series
        raise ValueError(f"No series named '{series_name}' exists in Chart '{self._name}'")

    def _get_series_index(self, series_name: str) -> Tuple[int, SeriesBase]:
        for i, series in enumerate(self._series):
            if series._name == series_name:
                return i, series
        raise ValueError(f"No series named '{series_name}' exists in Chart '{self._name}'")

    @overload
    def with_formula(self, /, formula: _Formula, **kwargs) -> "Chart":
        ...

    @overload
    def with_formula(self, /, name: str, **kwargs) -> "Chart":
        ...

    @_copy_with
    def with_formula(self, /, name: Optional[str] = None, formula: Optional[_Formula] = None, **kwargs) -> "Chart":
        if formula:
            if not isinstance(formula, _Formula):
                raise ValueError(f"'formula' value '{formula}' must be a Formula object")
            self._formulas.append(formula)
            return self
        if name:
            if not isinstance(name, str):
                raise ValueError(f"'name' value '{name}' must be a string")
            new_formula = _Formula(**kwargs)
            self._formulas.append(new_formula)
            return self
        raise ValueError("Either a 'name' for a new formula or an existing Formula 'formula' must be supplied")

    @_copy_with
    def with_trigger(
        self,
        expression: ThresholdFunction,
        trigger_name: str,
        severity: Optional[Union[AlertSeverityKind, str]] = None,
        channel_name: Optional[str] = None,
    ) -> "Chart":
        """Attaches a `Trigger` to your `Chart`. Your `Chart` may optionally have one `Trigger`.

        Parameters
        ----------
        expression
            Triggers are applied when a certain series is above or below a given value.
            The expression specifies the series, operand, and value as follows
                - the left-hand side of the expression must be a Series instance.
                - the operand must be `<` or `>`
                - the right-hand side must be an `int` or `float`
            Thus, if we have a Series instance `series1`, `expression=series1 > 0.5`
            will result in an alert when `series` is greater than 0.5.
        trigger_name
            The name for the new trigger.
        severity
            The severity of the trigger.
                - `critical`
                - `error`
                - `warning`
                - `info`
        channel_name
            The owner or email of the trigger.

        Returns
        -------
        Chart
            A copy of your chart with the new trigger.
        """
        if not isinstance(expression.lhs, SeriesBase):
            raise ValueError(f"Left hand side of expression '{expression.lhs}' must be a Series")
        threshold_position = ThresholdPosition.ABOVE if expression.operation == ">" else ThresholdPosition.BELOW
        trigger = Trigger(
            name=trigger_name,
            severity=AlertSeverityKind(severity.upper()),
            threshold_position=threshold_position,
            threshold_value=expression.rhs,
            series_name=expression.lhs._name,
            channel_name=channel_name,
        )
        self._trigger = trigger
        return self

    @_copy_with
    def with_feature_link(self, feature: Any) -> "Chart":
        """Explicitly link a Chart to a feature.
        This chart will then be visible on the webpage for this feature.
        Charts may only be linked to one entity.

        Parameters
        ----------
        feature
            A Chalk feature

        Returns
        -------
        Chart
            A copy of your chart linked to the feature.
        """
        self._entity_kind = ChartLinkKind.feature
        self._entity_id = str(feature) if isinstance(feature, FeatureWrapper) else feature
        return self

    @_copy_with
    def with_resolver_link(self, resolver: ResolverProtocol) -> "Chart":
        """Explicitly link a chart to a resolver.
        This chart will then be visible on the webpage for this resolver.
        Charts may only be linked to one entity.

        Parameters
        ----------
        resolver
            A Chalk resolver

        Returns
        -------
        Chart
            A copy of your chart linked to the resolver.
        """
        self._entity_kind = ChartLinkKind.resolver
        self._entity_id = resolver.fqn if isinstance(resolver, Resolver) else resolver
        return self

    @_copy_with
    def with_query_link(self, query_name: str) -> "Chart":
        """Explicitly link a chart to a query.
        This chart will then be visible on the webpage for this query.
        Charts may only be linked to one entity.

        Parameters
        ----------
        query_name
            A name of a Chalk query

        Returns
        -------
        Chart
            A copy of your chart linked to the query.
        """
        self._entity_kind = ChartLinkKind.query
        self._entity_id = query_name
        return self

    def keep(self) -> "Chart":
        """Designates that this chart and all of its descendants will be registered.

        Returns
        -------
        Chart
            The same chart.
        """
        self._keep = True
        return self

    def __str__(self) -> str:
        return f"Chart(name='{self._name}')"

    def __getitem__(self, key: str) -> Union[SeriesBase, _Formula]:
        """Retrieve a series or formula by name from a chart

        Parameters
        ----------
        key
            The name of the series or formula to retrieve.

        Returns
        -------
        The series or formula with the given name.

        """
        for series in self._series:
            if series._name == key:
                return series
        for formula in self._formulas:
            if formula._name == key:
                return formula
        raise ValueError(f"No series or formula named '{key}' exists in Chart {self._name}")

    def __eq__(self, obj):
        return hash(self) == hash(obj)

    def __hash__(self):
        name = self._name if self._name else "."

        window_period = self._window_period if self._window_period else "."
        chart_string = f"chart.{name}.{window_period}"

        series_hash = ".".join(sorted([str(hash(series)) for series in self._series]))
        formulas = ".".join(sorted([str(hash(formula)) for formula in self._formulas]))
        trigger = str(hash(self._trigger)) if self._trigger else "."

        return zlib.crc32(chart_string.encode() + series_hash.encode() + formulas.encode() + trigger.encode())
